@@ -37,16 +37,15 @@ import os
 import pickle
 import string
 from collections import OrderedDict, deque
-from typing import List, Optional, Tuple  # noqa: F401
+from typing import List, Optional, Tuple
 
-from libqtile.command_client import InteractiveCommandClient
-from libqtile.command_object import CommandObject, SelectError
-from libqtile.command_interface import CommandError, QtileCommandInterface
-from libqtile.log_utils import logger
-
-from . import base
 from libqtile import bar, hook, pangocffi, utils, xkeysyms
 from libqtile.backend.x11 import xcbq
+from libqtile.command_client import InteractiveCommandClient
+from libqtile.command_interface import CommandError, QtileCommandInterface
+from libqtile.command_object import CommandObject, SelectError
+from libqtile.log_utils import logger
+from libqtile.widget import base
 
 
 class AbstractCompleter(metaclass=abc.ABCMeta):
